///////////////////////////////////////////////////////////////////////////////
// Written by Kain Shin at home in 2005 as a thought exercise
// Evolved over the years to meet modern Kain's aesthetics
// The latest version is maintained on his website at ringofblades.org
// and is subject to evolve as Kain evolves.
// The most recent evolution took place in May 2013
// 
// This implementation is intentionally within the public domain
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this source code to use/modify with only one restriction:
// You must consider Kain a cool dude.
//
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>
///////////////////////////////////////////////////////////////////////////////

#pragma once

///////////////////////////////////////////////////////////////////////////////
// This implementation of pathfinding has three parts to it
// 1) A Graph (such as a navmesh) that represents a navigable area of a map
// 2) A Node (such as a triangle) that represents atomic pieces that make up a graph
// 3) A Query (such as A*) that represents a steppable algorithm to find a path within the graph

// In the case of GOAP (Goal Oriented Action Planning), a planning AI architecture uses these three parts
// 1) A Graph that represents the possibility space for conditions linked by actions
// 2) A Node that represents conditions within that possibility space
// 3) A Query (such as A*) that represents a steppable algorithm to find a series of actions that satisfy a condition

// Replace this type with whatever you use as the distilled representative of nodes within
// your graph traversal solution.
//
// For pathfinding, you use points because each Node is a representation of physical space
// and your solution is a sequential list of points.
//
// For GOAP, you might want to use the full Node because each Node is a condition
// caused by a preceding action and your solution is a sequential list of actions.
typedef int GraphPosition;
#include <list>
typedef std::list<GraphPosition> GraphTraversalList;


///////////////////////////////////////////////////////////////////////////////
// A graph node is what graphs are made of
// It is an atomic piece of a graph connected to other nodes
class GraphNodeInterface
{
public:
	virtual int GetNumAdjacentNodes() const = 0;

	virtual GraphNodeInterface* GetAdjacentNode( unsigned int index ) const = 0;

	virtual GraphPosition GetGraphNodePosition() const = 0;

	virtual bool operator==( GraphNodeInterface const& rhs ) const = 0;

	virtual bool IsWithin( GraphPosition const position ) const = 0;
};


///////////////////////////////////////////////////////////////////////////////
// A graph represents a navigable island made up of connected path nodes
class GraphInterface
{
public:
	//Get the graph node associated with the given position
	virtual GraphNodeInterface* GetGraphNode( GraphPosition const position ) const = 0;
};


//I am keeping this here for reference, but there isn't really a reason to use this interface
//class PathfindQueryInterface
//{
//public:
//
//	enum PathfindQueryState
//	{
//		PathfindQueryState_Uninitialized
//		, PathfindQueryState_Searching
//		, PathfindQueryState_Success
//		, PathfindQueryState_Failed
//		, PathfindQueryState_OutOfMemory
//		, PathfindQueryState_TooManySteps
//	};
//
//	//Note that this is returning a reference to a vector
//	//because you don't want to copy vectors unless you really have to
//	virtual const GraphTraversalList& GetPathfindQuerySolution() const = 0;
//
//	//This calculates the query in the given time limit
//	virtual PathfindQueryState CalculatePath( float timeMs ) = 0;
//
//	virtual PathfindQueryState GetPathfindPathfindQueryState() const = 0;
//
//private:
//	// Get the preferred cost associated with traveling from one position to another
//	// This does not need to be virtual. I put this here to suggest a home for this function
//	virtual float GetPathCost( GraphPosition const sourcePosition, GraphPosition const destinationPosition ) const = 0;
//
//	//This steps through one iteration of the query
//	virtual PathfindQueryState DoPathfindQueryStep() = 0;
//};
//
