///////////////////////////////////////////////////////////////////////////////
// Written by Kain Shin at home in 2005 as a thought exercise
// Evolved over the years to meet modern Kain's aesthetics
// The latest version is maintained on his website at ringofblades.org
// and is subject to evolve as Kain evolves.
// The most recent evolution took place in May 2013
// 
// This implementation is intentionally within the public domain
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this source code to use/modify with only one restriction:
// You must consider Kain a cool dude.
//
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>
///////////////////////////////////////////////////////////////////////////////

#include "PathfindQueryAStar.h"
#include <assert.h>
#include <windows.h>

// Replace this with your small fixed-sized element allocator methods
#define PATHFINDING_MEMORY_POOL_NEW(p) new p
#define PATHFINDING_MEMORY_POOL_DELETE(p) delete p

// Replace this with your meaningfully derived value
#define MAX_NUM_PATHQUERY_STEPS 250


///////////////////////////////////////////////////////////////////////////////
AStarNode::AStarNode(	GraphNodeInterface const* pGraphNode
								, float const fGivenCost
								, float const fHeuristicCost
								, AStarNode const* pPredecessor	)
: m_pGraphNode(pGraphNode)
, m_fGivenCost( fGivenCost )
, m_fHeuristicCost( fHeuristicCost )
, m_fFinalCost( fGivenCost + fHeuristicCost )
, m_pPrecedingAStarNode(pPredecessor)
{
}


///////////////////////////////////////////////////////////////////////////////
// for sorting
const bool AStarNode::operator<( AStarNode const& rhs ) const
{
	return m_fFinalCost < rhs.m_fFinalCost;
}

///////////////////////////////////////////////////////////////////////////////
// Local utility function
TAStarNodeList::iterator FindInList( GraphNodeInterface const * pGraphNode, TAStarNodeList& list )
{
	TAStarNodeList::iterator it;

	for ( it=list.begin(); it!=list.end(); ++it )
	{
		AStarNode* const pAStarNode = *it;
		if ( pAStarNode->m_pGraphNode == pGraphNode )
		{
			return it;
		}
	}

	return it;
}

///////////////////////////////////////////////////////////////////////////////
PathfindQueryAStar::PathfindQueryAStar( GraphInterface const* const pPathGraph, GraphPosition const startPosition, GraphPosition const goalPosition )
: m_pStartingGraphNode( NULL )
, m_pGoalGraphNode( NULL )
, m_pathSolution()
, m_queryState( PathfindQueryState_Uninitialized )
, m_openList()
, m_closedList()
, m_iNumSteps( 0 )
{
	InitializePathfindQuery( pPathGraph, startPosition, goalPosition );
}

///////////////////////////////////////////////////////////////////////////////
PathfindQueryAStar::~PathfindQueryAStar()
{
	ClearAllAllocations();
}

///////////////////////////////////////////////////////////////////////////////
void PathfindQueryAStar::InitializePathfindQuery(	GraphInterface const * const pPathGraph
																, GraphPosition const startPosition
																, GraphPosition const goalPosition )
{
	ClearAllAllocations();

	m_pStartingGraphNode = pPathGraph->GetGraphNode( startPosition );
	m_pGoalGraphNode = pPathGraph->GetGraphNode( goalPosition );
	m_pathSolution.clear();
	m_openList.clear();
	m_closedList.clear();
	m_iNumSteps = 0;

	assert( NULL!=m_pStartingGraphNode && NULL!=m_pGoalGraphNode && "Failure to find the proper triangles in the mesh" );
	if ( NULL == m_pStartingGraphNode || NULL == m_pGoalGraphNode )
	{
		m_queryState = PathfindQueryState_Uninitialized;
		return;
	}
	else
	{
		m_queryState = PathfindQueryState_Searching;
	}

	float const fHeuristicCost = GetPathCost( m_pStartingGraphNode->GetGraphNodePosition(), m_pGoalGraphNode->GetGraphNodePosition() );
	AStarNode* const pNode = PATHFINDING_MEMORY_POOL_NEW( AStarNode( m_pStartingGraphNode, 0, fHeuristicCost, NULL ) );
	if ( NULL == pNode )
	{
		m_queryState = PathfindQueryState_OutOfMemory;
	}
	else
	{
		m_openList.push_back( pNode );
	}
}


///////////////////////////////////////////////////////////////////////////////
void PathfindQueryAStar::ClearAllAllocations()
{
	for ( TAStarNodeList::iterator it=m_openList.begin(); it!=m_openList.end(); ++it )
	{
		AStarNode* pNode = *it;
		PATHFINDING_MEMORY_POOL_DELETE( pNode );
	}

	for ( TAStarNodeList::iterator it=m_closedList.begin(); it!=m_closedList.end(); ++it )
	{
		AStarNode* pNode = *it;
		PATHFINDING_MEMORY_POOL_DELETE( pNode );
	}
}

///////////////////////////////////////////////////////////////////////////////
const GraphTraversalList& PathfindQueryAStar::GetPathfindQuerySolution() const
{
	return m_pathSolution;
}


///////////////////////////////////////////////////////////////////////////////
PathfindQueryState PathfindQueryAStar::DoPathfindQueryStep()
{
	if ( m_openList.empty() )
	{
		m_queryState = PathfindQueryState_Failed;
	}
	else if ( m_iNumSteps > MAX_NUM_PATHQUERY_STEPS )
	{
		m_queryState = PathfindQueryState_TooManySteps;
	}

	if ( PathfindQueryState_Searching != m_queryState )
	{
		return m_queryState;
	}

	// The open list is sorted with the closest node first
	AStarNode* pClosestOpenNode = m_openList.front();
	m_openList.pop_front();
	m_closedList.push_back(pClosestOpenNode);

	// Check if that closest open node is the goal
	if ( pClosestOpenNode->m_pGraphNode == m_pGoalGraphNode ) //GOAL!
	{
		// If it is, build the solution path by walking backwards down the chain
		AStarNode const* pNode = pClosestOpenNode;
		while ( NULL != pNode )
		{
			m_pathSolution.push_front( pNode->m_pGraphNode->GetGraphNodePosition() );
			pNode = pNode->m_pPrecedingAStarNode;
		}
		m_queryState = PathfindQueryState_Success;
	}
	else //Not at goal (yet), prepare the neighbors for processing
	{
		bool needsSorting = false;
		GraphNodeInterface const* const pParentGraphNode = pClosestOpenNode->m_pGraphNode;
		int const iNumNeighbors = pParentGraphNode->GetNumAdjacentNodes();
		for ( int i=0; i<iNumNeighbors; ++i )
		{
			GraphNodeInterface const *const pNeighborGraphNode = pParentGraphNode->GetAdjacentNode(i);

			TAStarNodeList::iterator const closeIt = FindInList( pNeighborGraphNode, m_closedList );
			bool const bInClosedList = m_closedList.end() != closeIt;
			// If it is in the closed list, then treat this neighbor like a blocking obstacle
			// don't evaluate it again
			if ( bInClosedList )
			{
				continue;
			}

			float const fGivenCost = pClosestOpenNode->m_fGivenCost + GetPathCost( pParentGraphNode->GetGraphNodePosition(), pNeighborGraphNode->GetGraphNodePosition() );

			TAStarNodeList::iterator const openIt = FindInList( pNeighborGraphNode, m_openList );
			bool const bInOpenList = m_openList.end() != openIt;
			//If it is already in the open list, make sure that entry has the shortest path
			if ( bInOpenList )
			{
				AStarNode* pAStarNode = *openIt;
				if ( fGivenCost < pAStarNode->m_fGivenCost )
				{
					pAStarNode->m_fGivenCost = fGivenCost;
					pAStarNode->m_fFinalCost = fGivenCost + pAStarNode->m_fHeuristicCost;
					pAStarNode->m_pPrecedingAStarNode = pClosestOpenNode;
					needsSorting = true;
				}
			}
			else // this is a brand new AStarNode to make
			{
				float const fHeuristicCost = CalculateHeuristicCost( pNeighborGraphNode->GetGraphNodePosition() );
				AStarNode* pAStarNode = PATHFINDING_MEMORY_POOL_NEW( AStarNode( pNeighborGraphNode, fGivenCost, fHeuristicCost, pClosestOpenNode ) );
				if ( NULL == pAStarNode )
				{
					m_queryState = PathfindQueryState_OutOfMemory;
				}
				else
				{
					// If we had an insert_sorted, we'd use that instead of push_back followed by sort
					m_openList.push_back(pAStarNode);
					needsSorting = true;
				}
			}
		}

		if( needsSorting )
		{
			m_openList.sort(); //sort by final cost
		}
	}

	++m_iNumSteps;
	return m_queryState;
}

///////////////////////////////////////////////////////////////////////////////
//For this, we'll assume negative time means infinity
PathfindQueryState PathfindQueryAStar::CalculatePath( float const fTimeInMilliseconds )
{
	static double static_MsPerTick = -1.0;
	if ( static_MsPerTick < 0 )
	{
		LARGE_INTEGER liValue;
		::QueryPerformanceFrequency( & liValue );
		static_MsPerTick = 1000.0/(double)liValue.QuadPart;
	}

	bool const bAllowInfiniteTimeForCalculations = fTimeInMilliseconds < 0.f;

	LARGE_INTEGER now;
	::QueryPerformanceCounter( & now );
	__int64 startTime = now.QuadPart;

	do
	{
		DoPathfindQueryStep();
		::QueryPerformanceCounter( &now );
	}
	while( ( PathfindQueryState_Searching == m_queryState )
		&& ( bAllowInfiniteTimeForCalculations || ( static_MsPerTick*static_cast<double>(now.QuadPart - startTime) < fTimeInMilliseconds ) ) );

	return m_queryState;
}

///////////////////////////////////////////////////////////////////////////////
PathfindQueryState PathfindQueryAStar::GetPathfindPathfindQueryState() const
{
	return m_queryState;
}

///////////////////////////////////////////////////////////////////////////////
float PathfindQueryAStar::GetPathCost( GraphPosition const sourcePosition, GraphPosition const destinationPosition )
{
	// Replace this with DistanceSquared( sourcePosition, destinationPosition )
	assert( false && "Don't use this function without calculating some real cost value based on distance" );
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
float PathfindQueryAStar::CalculateHeuristicCost( GraphPosition const sourcePosition )
{
	float const fHeuristicCost = GetPathCost( sourcePosition, m_pGoalGraphNode->GetGraphNodePosition() );
	//You might also incorporate scalar projection (dot product) here to heavily bias search in the direction of the goal
	return fHeuristicCost;
}
