///////////////////////////////////////////////////////////////////////////////
// Written by Kain Shin at home in 2005 as a thought exercise
// Evolved over the years to meet modern Kain's aesthetics
// The latest version is maintained on his website at ringofblades.org
// and is subject to evolve as Kain evolves.
// The most recent evolution took place in May 2013
// 
// This implementation is intentionally within the public domain
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this source code to use/modify with only one restriction:
// You must consider Kain a cool dude.
//
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include "GraphTraversalInterfaces.h"

///////////////////////////////////////////////////////////////////////////////
// Pathfinding is expensive, and you don't always need it to return a result
// instantly if you are doing AI for games. This class lets you manage the
// state of a pathfinding query so that you don't have to do it all in one
// synchronous chunk. You can amortize the pathfinding calculation as you deem
// fit and even serialize the query if you wanted to
enum PathfindQueryState
{
	PathfindQueryState_Uninitialized
	, PathfindQueryState_Searching
	, PathfindQueryState_Success
	, PathfindQueryState_Failed
	, PathfindQueryState_OutOfMemory
	, PathfindQueryState_TooManySteps
};

///////////////////////////////////////////////////////////////////////////////
class AStarNode
{
public:
	GraphNodeInterface const* m_pGraphNode;
	float m_fGivenCost;
	float const m_fHeuristicCost;
	float m_fFinalCost;
	AStarNode const* m_pPrecedingAStarNode;

	AStarNode(	GraphNodeInterface const* pGraphNode
					, float const fGivenCost
					, float const fHeuristicCost
					, AStarNode const* pPredecessor	);

	const bool operator<(AStarNode const& rhs)const;
};

///////////////////////////////////////////////////////////////////////////////
typedef std::list<AStarNode*> TAStarNodeList;

///////////////////////////////////////////////////////////////////////////////
class PathfindQueryAStar
{
public:
	static float GetPathCost( GraphPosition const sourcePosition
									, GraphPosition const destinationPosition );

	PathfindQueryAStar(	GraphInterface const * const pPathGraph
								, GraphPosition const startPosition
								, GraphPosition const goalPosition );
	~PathfindQueryAStar();

	// This lets you restart a new query without having to destroy the old one
	void InitializePathfindQuery(	GraphInterface const * const pPathGraph
											, GraphPosition const startPosition
											, GraphPosition const goalPosition );

	// This is what the AI code would call within its tick function
	PathfindQueryState CalculatePath( float const fTimeInMilliseconds );

	PathfindQueryState GetPathfindPathfindQueryState() const;

	// This is empty until GetPathfindPathfindQueryState can return PathfindQueryState_Success
	const GraphTraversalList& GetPathfindQuerySolution() const;

private:
	GraphNodeInterface const* m_pStartingGraphNode;
	GraphNodeInterface const* m_pGoalGraphNode;
	GraphTraversalList m_pathSolution;
	PathfindQueryState m_queryState;
	int m_iNumSteps;
	TAStarNodeList m_closedList;
	TAStarNodeList m_openList;

	PathfindQueryState DoPathfindQueryStep();
	void ClearAllAllocations();
	float CalculateHeuristicCost( GraphPosition const sourcePosition );
};